"""BOX is a Zeroton meant to be used by the AttriBox class as a
placeholder for the box instance itself. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import Zeroton


class BOX(Zeroton):
  """BOX is a Zeroton meant to be used by the AttriBox class as a
  placeholder for the box instance itself. """
