"""ATTR is a Zeroton meant to be used by the AttriBox class as a
placeholder for the box class itself. This would be AttriBox or a subclass
of it."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import Zeroton


class ATTR(Zeroton):
  """ATTR is a Zeroton meant to be used by the AttriBox class as a
  placeholder for the box class itself. This would be AttriBox or a subclass
  of it."""
