"""The 'worktoy.parse' package provides tools for parsing data."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._maybe import maybe
from ._maybe_type import maybeType
