"""THIS is a Zeroton meant to be used by the AttriBox class as a
placeholder for the instance of the class owning the AttriBox. This allows
the inner object managed by the AttriBox instance to include the owning
instance during class creation. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import Zeroton


class THIS(Zeroton):
  """THIS is a Zeroton serving as placeholder for the yet to be created
  instance owning the AttriBox instance. """
